@@ -147,14 +147,10 @@ fk_truncate_illegal_if_parent(THD *thd, TABLE *table)
   /* Loop over the set of foreign keys for which this table is a parent. */
   while ((fk_info= it++))
   {
-    if (lex_string_cmp(system_charset_info, fk_info->referenced_db,
-                       &table->s->db) ||
-        lex_string_cmp(system_charset_info, fk_info->referenced_table,
-                       &table->s->table_name) ||
-        lex_string_cmp(system_charset_info, fk_info->foreign_db,
-                       &table->s->db) ||
-        lex_string_cmp(system_charset_info, fk_info->foreign_table,
-                       &table->s->table_name))
+    if (!table->s->db.streq(*fk_info->referenced_db) ||
+        !table->s->table_name.streq(*fk_info->referenced_table) ||
+        !table->s->db.streq(*fk_info->foreign_db) ||
+        !table->s->table_name.streq(*fk_info->foreign_table))
       break;
   }
 
@@ -482,6 +478,14 @@ bool Sql_cmd_truncate_table::truncate_table(THD *thd, TABLE_LIST *table_ref)
     if (lock_table(thd, table_ref, &hton_can_recreate))
       DBUG_RETURN(TRUE);
 
+    /*
+      This is mainly here for truncate_notembedded.test, but it is still
+      useful to check killed after we got the lock
+    */
+
+    if (thd->killed)
+      DBUG_RETURN(TRUE);
+
     if (hton_can_recreate)
     {
      /*
@@ -511,7 +515,7 @@ bool Sql_cmd_truncate_table::truncate_table(THD *thd, TABLE_LIST *table_ref)
            (HTON_REQUIRES_CLOSE_AFTER_TRUNCATE |
             HTON_TRUNCATE_REQUIRES_EXCLUSIVE_USE)))
       {
-        thd->locked_tables_list.mark_table_for_reopen(thd, table_ref->table);
+        thd->locked_tables_list.mark_table_for_reopen(table_ref->table);
         if (unlikely(thd->locked_tables_list.reopen_tables(thd, false)))
           thd->locked_tables_list.unlink_all_closed_tables(thd, NULL, 0);
       }
